#include <vector>
#include <fstream>
#include <iostream>

#include "CrossValidator.h"
#include "message.h"
#include "../mpi_wrapper.h"

namespace lib4neuro {
    LIB4NEURO_API CrossValidator::CrossValidator(LearningMethod* optimizer,
                                                 ErrorFunction* ef) {
        this->optimizer = optimizer;
        this->ef        = *(new std::vector<ErrorFunction*>);
        this->ef.emplace_back(ef);
    }

    LIB4NEURO_API void CrossValidator::run_k_fold_test(unsigned int k,
                                                       unsigned int tests_number,
                                                       std::ofstream* results_file_path) {
        //TODO do not duplicate code - write in a more elegant way
        //NeuralNetwork* net = this->ef->get_network_instance();

//        double cv_err_sum = 0;
        std::vector<double> cv_err_sum(this->ef.size(), 0);

        for (unsigned int i = 0; i < tests_number; i++) {
            COUT_INFO("Cross-validation run " << i + 1 << std::endl);

            this->ef.at(0)->divide_data_train_test(1.0 / k);
            int ntot_train = this->ef.at(0)->get_n_data_set();
            int ntot_test = this->ef.at(0)->get_n_test_data_set();
            MPI_Allreduce( MPI_IN_PLACE, &ntot_train, 1, MPI_INT, MPI_SUM, lib4neuro::mpi_active_comm );
            MPI_Allreduce( MPI_IN_PLACE, &ntot_test, 1, MPI_INT, MPI_SUM, lib4neuro::mpi_active_comm );

            COUT_INFO("Number of train data points: " << ntot_train << std::endl);
            COUT_INFO("Number of test data points: " << ntot_test << std::endl);
            this->ef.at(0)->randomize_parameters(1.0 / ef.at(0)->get_dimension());

            COUT_INFO("Starting optimization...");
            this->optimizer->optimize( *this->ef.at(0) );
            COUT_INFO("Finished");

            /* Error evaluation and writing */
            double err;
            for(size_t err_func_i = 0; err_func_i < this->ef.size(); err_func_i++) {
                if(!results_file_path) {
                    err = this->ef.at(err_func_i)->eval_on_test_data();
                } else {
                    err = this->ef.at(err_func_i)->eval_on_test_data(results_file_path);
                }
                cv_err_sum.at(err_func_i) += err;
                COUT_INFO("CV error - error function " << err_func_i << ": (run " << i + 1 << "): " << err << std::endl << std::endl);
            }
            this->ef.at(0)->return_full_data_set_for_training();
        }

        for(size_t err_func_i = 0; err_func_i < this->ef.size(); err_func_i++) {
            COUT_INFO("CV error (error function " << err_func_i << ") mean: " << cv_err_sum.at(err_func_i) / tests_number << std::endl);
            *results_file_path << "CV error (error function " << err_func_i << ") mean: " << cv_err_sum.at(err_func_i) / tests_number << std::endl;
        }
    }

    LIB4NEURO_API void CrossValidator::run_k_fold_test(unsigned int k,
                                                       unsigned int tests_number,
                                                       std::string results_file_path) {

        std::ofstream ofs(results_file_path, std::ios::in);
        this->run_k_fold_test(k, tests_number, &ofs);
        ofs.close();

//        double cv_err_sum = 0;
//
//        for (unsigned int i = 0; i < tests_number; i++) {
//            COUT_INFO("Cross-validation run " << i + 1 << std::endl);
//
//            this->ef->divide_data_train_test(1.0 / k);
//			int ntot_train = this->ef->get_n_data_set();
//			int ntot_test = this->ef->get_n_test_data_set();
//			MPI_Allreduce( MPI_IN_PLACE, &ntot_train, 1, MPI_INT, MPI_SUM, lib4neuro::mpi_active_comm );
//			MPI_Allreduce( MPI_IN_PLACE, &ntot_test, 1, MPI_INT, MPI_SUM, lib4neuro::mpi_active_comm );
//
//            COUT_INFO("Number of train data points: " << ntot_train << std::endl);
//            COUT_INFO("Number of test data points: " << ntot_test << std::endl);
//            this->ef->randomize_parameters(1.0 / ef->get_dimension());
//
//			COUT_INFO("Starting optimization...");
//            this->optimizer->optimize( *this->ef );
//			COUT_INFO("Finished");
//
//            /* Error evaluation and writing */
//            double err;
//            if (results_file_path.empty()) {
//                err = this->ef->eval_on_test_data();
//            } else {
//                err = this->ef->eval_on_test_data(results_file_path + "_cv" + std::to_string(i) + ".dat");
//            }
//            cv_err_sum += err;
//            COUT_INFO("CV error (run " << i + 1 << "): " << err << std::endl << std::endl);
//
//            this->ef->return_full_data_set_for_training();
//        }
//
//        COUT_INFO("CV error mean: " << cv_err_sum / tests_number << std::endl);
    }

    CrossValidator::CrossValidator(LearningMethod* optimizer,
                                   std::vector<ErrorFunction*>& ef) {
        this->optimizer = optimizer;
        this->ef = ef;
    }
}
